---@class List
local List = {}

local function reduce(list, f, init)
  local acc = init
  for i, v in ipairs(list) do
    acc = f(v, acc, i)
  end
  return acc
end

local function filter(list, predicate)
  return reduce(list, function(el, acc, i)
    if predicate(el) then
      table.insert(acc, el)
    end
    return acc
  end, List.create{})
end

local function shallow_copy(t)
  local t2 = {}
  for k, v in ipairs(t) do
    t2[k] = v
  end
  return t2
end

List.__index = List
setmetatable(List, {
  __call = function(self, ...)
    local args = {...}
    if #args == 0 then
      return List.create()
    end

    if type(args[1]) == "table" then
      return List.create(args[1])
    end
    return List.create(args)
  end
})

---Constructs a new list from the given table.
---@param from table?
---@return List
function List.create(from)
  from = from or {}
  local l = shallow_copy(from)
  return setmetatable(l, List)
end

function List:__tostring()
  local s = "List(" .. table.concat(self, ", ", 1, #self) .. ")"
  return s
end

---Appends value to the end of the list.
---@param value any
function List:push(value)
  if self:has(value) then return end
  table.insert(self, value)
end

---Removes the last element in the list and returns it.
---@return any
function List:pop()
  return table.remove(self, #self)
end

---Removes the first element in the list and returns it.
---@return any
function List:pop_front()
  return table.remove(self, 1)
end

---Reduce.
---@generic T
---@param f fun(element: any, accumulator: T, index: integer): T
---@param init T|nil initial value of accumulator
---@return T
function List:reduce(f, init)
  return reduce(self, f, init)
end

---Returns a new List of all elements of this list that match the predicate function.
---@param predicate fun(element: any): boolean
---@return List
function List:filter(predicate)
  return filter(self, predicate)
end

---Removes the value at the given index.
---@param idx integer
---@return any
function List:remove_at(idx)
  return table.remove(self, idx)
end

---Returns the index of value, if it exists in the list, -1 otherwise.
---@param value any
---@return integer
function List:find(value)
  local idx = -1
  for i, v in ipairs(self) do
    if v == value then
      idx = i
      break
    end
  end
  return idx
end

---Returns true if the value exists in the list.
---@param value any
---@return boolean
function List:has(value)
  return self:find(value) ~= -1
end

---Removes the value from the list, if it exists.
---@param value any
function List:remove_value(value)
  local idx = self:find(value)
  if idx ~= -1 then
    table.remove(self, idx)
  end
end

---Returns true if the list is empty.
---@return boolean
function List:is_empty()
  return #self == 0
end

---Returns a (shallow) copy of this list.
---@return List
function List:copy()
  return List.create(self)
end

---Returns a new copy of this list with the order of the elements shuffled around.
---@return List
function List:shuffled()
  -- https://gist.github.com/Uradamus/10323382
  local list = self:copy()
  
  for i = #list, 2, -1 do
    local j = math.random(i)
    list[i], list[j] = list[j], list[i]
  end
  
  return list
end

---Returns a sorted copy of this list.
---@param f? fun(a: any, b: any): boolean
---@return List
function List:sorted(f)
  local list = self:copy()
  table.sort(list, f)
  return list
end

return List