--- @alias vectorlike number[] | Vector3
--- @class Vector3
local Vector3 = {
  x = 0,
  y = 0,
  z = 0,

  _CLASS_ = "Vector3"
}

setmetatable(Vector3, Vector3)

---Returns true if the provided table can be coerced into a Vector3.
---@param t table
---@return boolean
local function is_weak_vector3(t)
  if type(t) ~= "table" then return false end

  local haskeys = t.x ~= nil and t.y ~= nil and t.z ~= nil
  if not haskeys then
    return type(t[1]) == "number" and type(t[2]) == "number" and type(t[3]) == "number"
  else
    return type(t.x) == "number" and type(t.y) == "number" and type(t.z) == "number"
  end
end

---Coerces t to a Vector3. If t is already a Vector3, returns it without copying.
---@param t vectorlike
---@param do_err boolean?
---@param def Vector3?
---@return boolean error, Vector3 result
local function coerce(t, do_err, def)
  do_err = do_err or true
  if do_err and not is_weak_vector3(t) then
    def = def or Vector3()
    error("Vector3: can not coerce t into Vector3. Returning " .. tostring(def))
    return true, def
  end
  if t["_CLASS_"] == "Vector3" then
    return false, t
  end

  return false, Vector3(t)
end

---Returns a Vector3 multiplied either component-wise (if b is a weak Vector3) or multiplies each component by b if b is a number.
---@param b number|vectorlike
---@return Vector3
function Vector3:__mul(b)
  if type(b) == "number" then
    return Vector3{
      x = self.x * b,
      y = self.y * b,
      z = self.z * b,
    }
  elseif type(b) == "table" and is_weak_vector3(b) then
    local other = Vector3(b)
    return Vector3{
      self.x * other.x,
      self.y * other.y,
      self.z * other.z,
    }
  end

  error("Vector3: other must be either a number or a Vector3-like table.")
  return Vector3()
end

---Returns a Vector3 divided either component-wise (if b is a weak Vector3) or divides each component by b if b is a number.
---@param b number|vectorlike
---@return Vector3
function Vector3:__div(b)
  if type(b) == "number" then
    return Vector3{
      x = self.x / b,
      y = self.y / b,
      z = self.z / b,
    }
  elseif type(b) == "table" and is_weak_vector3(b) then
    local other = Vector3(b)
    return Vector3{
      self.x / other.x,
      self.y / other.y,
      self.z / other.z,
    }
  end

  error("Vector3: other must be either a number or a Vector3-like table.")
  return Vector3()
end

---Returns a Vector3 with each component added together.
---@param b vectorlike
---@return Vector3
function Vector3:__add(b)
  if not is_weak_vector3(b) then
    error("Vector3: other must be a Vector3-like table.")
    return self
  end
  local other = Vector3(b)
  return Vector3{
    self.x + other.x,
    self.y + other.y,
    self.z + other.z,
  }
end

---Returns a Vector3 with b's components subtracted from this vector.
---@param b vectorlike
---@return Vector3
function Vector3:__sub(b)
  if not is_weak_vector3(b) then
    error("Vector3: other must be a Vector3-like table.")
    return self
  end
  local other = Vector3(b)
  return Vector3{
    self.x - other.x,
    self.y - other.y,
    self.z - other.z,
  }
end

---Returns a new Vector3 with each component being a negation of this vector.
---@return Vector3
function Vector3:__unm()
  return Vector3{
    -self.x,
    -self.y,
    -self.z,
  }
end

function Vector3:__call(...)
  local args = {...}
  local nv = {x = 0, y = 0, z = 0}
  setmetatable(nv, Vector3)
  if type(args[1]) == "number" then
    if #args == 1 then
      nv.x, nv.y, nv.z = args[1], args[1], args[1]
    else
      nv.x, nv.y, nv.z = args[1], args[2], args[3]
    end
  elseif type(args[1]) == "table" then
    if args[1].x ~= nil then
      nv.x, nv.y, nv.z = args[1].x, args[1].y, args[1].z
    else
      nv.x, nv.y, nv.z = args[1][1], args[1][2], args[1][3]
    end
  end
  return nv
end

function Vector3:__tostring()
  return "Vector3(" .. tostring(self.x) .. ", " .. tostring(self.y) .. ", " .. tostring(self.z) .. ")"
end

-- note: the < and <= operators of this class are component-wise rather than lexicographic (that is, htey are useful for bounds checking but are not suitable for sorting.)

function Vector3:__lt(b)
  local err, other = coerce(b, true)
  if err then return nil end
  return self.x < other.x and self.y < other.y and self.z < other.z
end

function Vector3:__le(b)
  local err, other = coerce(b, true)
  if err then return nil end
  return self.x <= other.x and self.y <= other.y and self.z <= other.z
end

local NUMK = {"x", "y", "z"}

function Vector3:__index(key)
  -- this allows constructs like `for i, v in ipairs(Vector3(3, 2, 1))` to iterate over components
  if type(key) == "number" then
    return rawget(self, NUMK[key])
  end

  return rawget(Vector3, key)
end

function Vector3:__newindex(key, value)
  if NUMK[key] then
    rawset(self, NUMK[key], value)
    return
  end

  rawset(self, key, value)
end
--------API--------

function Vector3:length_squared()
  local x2 = self.x * self.x
  local y2 = self.y * self.y
  local z2 = self.z * self.z

  return x2 + y2 + z2
end

function Vector3:length()
  return math.sqrt(self:length_squared())
end

function Vector3:normalized()
  local length = math.sqrt(self:length_squared())
  if length == 0 then
    return Vector3()
  end
  return Vector3(self.x / length, self.y / length, self.z / length)
end

---Returns the dot product of this vector with `with`.
---@param with vectorlike
---@return number
function Vector3:dot(with)
  local err, v2 = coerce(with)
  if err then
    return 0.0
  end

  return self.x * v2.x + self.y * v2.y + self.z * v2.z
end

---Returns the cross product of this vector with `with`.
---@param with vectorlike
---@return Vector3
function Vector3:cross(with)
  local err, v2 = coerce(with)
  if err then
    return v2
  end
  return Vector3 {
    self.y * v2.z - self.z * v2.y,
    self.z * v2.x - self.x * v2.z,
    self.x * v2.y - self.y * v2.x,
  }
end

---Returns a vector rotated about `axis` by `angle` radians
---@param axis vectorlike
---@param angle number
---@return Vector3
function Vector3:rotated(axis, angle)
  -- if not is_weak_vector3(axis) then
  --   error("Vector3: axis must be a Vector3-like table. Returning Vector3()")
  --   return Vector3()
  -- end
  local err, vaxis = coerce(axis)
  if err then
    return vaxis
  end

  vaxis = vaxis:normalized()
  local cosa = math.cos(angle)
  local sina = math.sin(angle)
  -- __mul is only defined for the left operand (table), numbers don't get metatables.
  -- as such, the ordering of operations here is specific
  return (self * cosa) + (vaxis * ((1 - cosa) * self:dot(vaxis))) + (vaxis:cross(self) * sina)
end

---In-place version of rotated.
---@param axis Vector3
---@param angle number
---@return Vector3
function Vector3:rotate(axis, angle)
  local cosa = math.cos(angle)
  local sina = math.sin(angle)

  local dot = self:dot(axis)
  local cross = axis:cross(self)

  self.x = self.x * cosa + axis.x * ((1 - cosa) * dot) + cross.x * sina
  self.y = self.y * cosa + axis.y * ((1 - cosa) * dot) + cross.y * sina
  self.z = self.z * cosa + axis.z * ((1 - cosa) * dot) + cross.z * sina

  return self
end

---In place set.
---@param x number
---@param y number
---@param z number
---@return Vector3
function Vector3:set(x, y, z)
  self.x, self.y, self.z = x, y, z
  return self
end

---In place set.
---@param t vectorlike
---@return Vector3
function Vector3:sett(t)
  self.x, self.y, self.z = t.x, t.y, t.z
  return self
end

---Returns a copy of this vector.
---@return Vector3
function Vector3:copy()
  return Vector3(self)
end

---Returns a new vector as a linear interpolation between self and b.
---@param b vectorlike
---@param t number
---@return Vector3
function Vector3:lerp(b, t)
  local err, w = coerce(b)
  if err then
    return w
  end

  return Vector3{
    util.lerp(self.x, w.x, t),
    util.lerp(self.y, w.y, t),
    util.lerp(self.z, w.z, t),
  }
end

function Vector3:direction_to(to)
  local err, other = coerce(to)
  if err then
    return 0.0
  end

  return Vector3(other.x - self.x, other.y - self.y, other.z - self.z):normalized()
end

---Returns the squared distance between this vector and to.
---@param to vectorlike
---@return number
function Vector3:distance_squared_to(to)
  local err, other = coerce(to)
  if err then
    return 0.0
  end

  return (other - self):length_squared()
end

---Returns the distance between this vector and to.
---@param to vectorlike
---@return number
function Vector3:distance_to(to)
  local err, other = coerce(to)
  if err then
    return 0.0
  end

  return (other - self):length()
end

---Returns a new vector with the Y discarded (0).
---@return Vector3
function Vector3:horizontal()
  return Vector3(self.x, 0.0, self.z)
end
---- CONSTANTS

---@type Vector3
Vector3.UP = Vector3(0, 1, 0)
---@type Vector3
Vector3.DOWN = Vector3(0, -1, 0)
---@type Vector3
Vector3.FORWARD = Vector3(0, 0, -1)
---@type Vector3
Vector3.BACK = Vector3(0, 0, 1)
---@type Vector3
Vector3.RIGHT = Vector3(1, 0, 0)
---@type Vector3
Vector3.LEFT = Vector3(-1, 0, 0)

-------------------
return Vector3
